/**
  *  \file client/dialogs/msgreceiver.cc
  *  \brief Message Receiver Selection
  *
  *  This module provides two dialogs to select message receivers.
  *  They are implemented in terms of a WPlayerSetSelector.
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/msgreceiver.h"
#include "client/widgets/helpwidget.h"
#include "client/widgets/setselect.h"
#include "game/spec.h"
#include "gfx/palette.h"
#include "ui/keystring.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/widgets/button.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/statictext.h"
#include "util/nls.h"

namespace {
    class WNewMessageReceiverWindow : public UIWindow {
        ObjectHolder h;
        WPlayerSetSelector& sel;
     public:
        WNewMessageReceiverWindow(WPlayerSetSelector& win, int numRevisable);
        void init(int numRevisable);
        void onUniversal();
    };
}

WNewMessageReceiverWindow::WNewMessageReceiverWindow(WPlayerSetSelector& sel, int numRevisable)
    : UIWindow(_("Send Message"), UIWindow::BLUE_WINDOW, COLOR_GRAY), sel(sel)
{
    init(numRevisable);
}

void
WNewMessageReceiverWindow::init(int numRevisable)
{
    setLayout(UIHBoxLayout::instance5);

    UIGroup& g1 = h.add(new UIGroup(UIVBoxLayout::instance5));
    g1.add(h.add(new UIStaticText(_("Choose receivers:"), UIStaticText::StaticTitle)));
    g1.add(UIFrameGroup::wrap(h, sel));

    g1.add(h.add(new UIButton(0, UIKeyString(_("Universal Message")))).addCall(this, &WNewMessageReceiverWindow::onUniversal));
    if (numRevisable != 0) {
        g1.add(h.add(new UIButton(nmsgRevise, UIKeyString(_("Revise messages")))).addStop());
    }
    add(g1);

    UIGroup& g2 = h.add(new UIGroup(UIVBoxLayout::instance5));
    g2.add(h.add(new UISpacer()));
    g2.add(h.add(new UIButton(nmsgConfirm,  _("OK"), SDLK_RETURN)).addStop());
    g2.add(h.add(new UIButton(nmsgCancel,   _("Cancel"), SDLK_ESCAPE)).addStop());
    g2.add(h.add(new UIButton(_("Help"), 'h')));
    add(g2);
    add(h.add(new WHelpWidget("pcc2:msgout")));

    pack();
}

void
WNewMessageReceiverWindow::onUniversal()
{
    sel.setSelectedPlayers(sel.getSelectedPlayers() | GPlayerSet::allUpTo(NUM_PLAYERS));
    stop(nmsgConfirm);
}

/******************** Public Interface Implementation ********************/

/** Set selectable players.
    \todo this could implement restrictions like VPA's DullMessages. */
static void
setSelectablePlayers(WPlayerSetSelector& sel)
{
    sel.setSelectablePlayers(GPlayerSet::allUpTo(NUM_OWNERS));
}

/** Choose receiver for a new message.
    \param sel [in/out] Receiver selector
    \param numRevisable [in] Number of revisable messages
    \retval nmsgConfirm User wants to send a new message
    \retval nmsgCancel User doesn't want to send a message
    \retval nmsgRevise User wants to revise existing messages */
TNewMessageResult
chooseNewMessageReceiver(WPlayerSetSelector& sel, const int numRevisable)
{
    setSelectablePlayers(sel);

    WNewMessageReceiverWindow win(sel, numRevisable);
    ui_root->centerWidget(win);
    TNewMessageResult result = TNewMessageResult(ui_root->runModal(win));
    if (result == nmsgConfirm && sel.getSelectedPlayers().empty()) {
        result = nmsgCancel;
    }
    return result;
}

/** Change receiver of an existing message.
    \param sel [in/out] Receiver selector
    \param title [in] Window title
    \param help [in] Help text Id */
bool
chooseMessageReceiver(WPlayerSetSelector& sel, const string_t title, const string_t help)
{
    setSelectablePlayers(sel);

    ObjectHolder h;

    // UIWindow/VBox
    //   WPlayerSetSelector
    //   UIButton "Universal"
    //   HBox
    //     UIButton "OK"
    //     UIButton "Cancel"
    //     UISpacer
    //     UIButton "Help"

    UIWindow& win = h.add(new UIWindow(title, UIWindow::BLUE_WINDOW, COLOR_GRAY));
    UIGroup& g3 = h.add(new UIGroup(UIHBoxLayout::instance5));

    win.setLayout(UIVBoxLayout::instance5);
    win.add(sel);
    win.add(h.add(new UIButton(100, UIKeyString(_("Universal Message"))))
            .addCall(&sel, &WPlayerSetSelector::toggleUniversal));
    win.add(g3);
    win.add(h.add(new WHelpWidget(help)));

    g3.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    g3.add(h.add(new UIButton(0, _("Cancel"), SDLK_RETURN)).addStop());
    g3.add(h.add(new UISpacer()));
    g3.add(h.add(new UIButton(_("Help"), 'h')));

    win.pack();
    ui_root->moveWidgetToEdge(win, 2, 2, 10);
    ui_root->add(win);
    while (1) {
        bool rv = ui_root->run();
        if (!rv || !sel.getSelectedPlayers().empty()) {
            ui_root->close(win);
            return rv;
        }
        messageBox(_("You must choose one or more receivers for your message."), _("Edit Message"));
    }
}
