/**
  *  \file client/screens.cc
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens.h"
#include "cpluslib/assert.h"
#include "game/ship.h"
#include "game/planet.h"
#include "game/global.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "game/fleet.h"

/** Create blank screen Id.
    \post !isValid */
WScreenId::WScreenId()
    : screen_id(NullScreen), x(0), y(0)
{ }

/** Create screen Id for an object.
    \param screen_id Screen identifier
    \param id        Selected object on that screen. Must be valid (not checked!)
                     or zero. */
WScreenId::WScreenId(ScreenId screen_id, int id)
    : screen_id(screen_id), x(id), y(0)
{
    // FIXME: add this when all code is converted
    // ASSERT(screen_id != Starchart);
}

/** Create screen Id for map location.
    \param pt Map location */
WScreenId::WScreenId(GPoint pt)
    : screen_id(Starchart), x(pt.x), y(pt.y)
{ }

/** Get object Id. */
int
WScreenId::getId() const
{
    ASSERT(screen_id != Starchart);
    ASSERT(screen_id != NullScreen);
    return x;
}

/** Get starchart position.
    \pre getScreenId() == Starchart */
GPoint
WScreenId::getPos() const
{
    ASSERT(screen_id == Starchart);
    return GPoint(x, y);
}

bool
WScreenId::isValid() const
{
    switch (screen_id) {
     case RaceScreen:
     case Starchart:
        return true;

     case ShipScreen:
     case ShipTaskScreen:
        return global_ship_type.isExistant(x);

     case PlanetScreen:
     case PlanetTaskScreen:
        return global_planet_type.isExistant(x);

     case HistoryScreen:
        return global_histship_type.isExistant(x);

     case BaseScreen:
     case BaseTaskScreen:
        return global_base_type.isExistant(x);

     case FleetScreen:
        return global_fleet_type.isExistant(x);

     default:
        return false;
    }
}

string_t
WScreenId::toString()
{
    switch (screen_id) {
     case RaceScreen:
        return _("Race Screen");
     case Starchart:
        return format(_("Starchart (%d,%d)"), x, y);
     case ShipScreen:
        return getDisplayedTurn().getCurrentUniverse().getShip(x).getFullName();
     case PlanetScreen:
        return getDisplayedTurn().getCurrentUniverse().getPlanet(x).getFullName();
     case BaseScreen:
        return format(_("Starbase #%d: %s"), x, getDisplayedTurn().getCurrentUniverse().getPlanet(x).getName());
     case HistoryScreen:
        return format(_("Ship History %d: %s"), x, getDisplayedTurn().getCurrentUniverse().getShip(x).getName());
     case FleetScreen:
        return getFleetTitle(getDisplayedTurn().getCurrentUniverse(), x);
     case ShipTaskScreen:
        return format(_("Ship Task #%d: %s"), x, getDisplayedTurn().getCurrentUniverse().getShip(x).getName());
     case PlanetTaskScreen:
        return format(_("Planet Task #%d: %s"), x, getDisplayedTurn().getCurrentUniverse().getPlanet(x).getName());
     case BaseTaskScreen:
        return format(_("Starbase Task #%d: %s"), x, getDisplayedTurn().getCurrentUniverse().getPlanet(x).getName());
     default:
        return format("<%d,%d,%d>", int(screen_id), x, y);
    }
}
