/**
  *  \file client/screens/fleetscreen.cc
  *  \brief Fleet Screen
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens/fleetscreen.h"
#include "client/tiles/fleetoverview.h"
#include "game/global.h"
#include "ui/msgbox.h"
#include "ui/window.h"
#include "client/dlg-objsel.h"
#include "game/ship.h"
#include "util/nls.h"
#include "game/fleet.h"
#include "ui/uiglobal.h"
#include "ui/commands.h"
#include "client/cscreen.h"
#include "client/wscanchart.h"
#include "int/intglobal.h"
#include "int/if/shipif.h"
#include "gfx/gfxcontext.h"
#include "gfx/palette.h"
#include "game/host.h"

namespace {
    class WFleetScannerChart : public WScannerChartWidget {
        GPoint member_pos;
     public:
        WFleetScannerChart();
        void lockQueryLocation(bool& hyperjumping, int& current_gravity_pid);
        int32_t lockQueryDistance(GPoint pt);
        void drawScanner(GfxCanvas& can);
        void drawPre(GfxCanvas& can);
        void setMemberPosition(GPoint pos);
    };

    class WFleetScreen : public WControlScreen {
        WFleetScannerChart chart;
        SignalConnection conn_member;
     public:
        WFleetScreen();
        ~WFleetScreen();
        void update();
        void setScriptContext();
        void onMemberChange();
    };
}

int
chooseFleet()
{
    if (global_fleet_type.isEmpty()) {
        messageBox(_("You have not yet established a fleet. "
                     "To start a fleet, use the \"Start New Fleet\" command on a ship's [F10] menu."),
                   _("Fleet"));
        return 0;
    } else if (global_fleet_type.isUnit()) {
        return global_fleet_type.getCurrent();
    } else {
        GAliasObjectType aot(global_fleet_type);
        WFleetOverviewTile bot(aot);
        WObjectSelectionDialog dlg(_("Select Fleet"), "pcc2:unitsel", aot);
        dlg.addWidget(bot);
        dlg.pack();
        ui_root->centerWidget(dlg);
        if (ui_root->runModal(dlg) == cm_Enter)
            return aot.getCurrent();
        else
            return 0;
    }
}

/*************************** WFleetScannerChart **************************/

WFleetScannerChart::WFleetScannerChart()
    : member_pos(0, 0)
{ }

void
WFleetScannerChart::lockQueryLocation(bool& hyperjumping, int& current_gravity_pid)
{
    // FIXME: this routine also in WShipScannerChartWidget
    if (GShip* sh = dynamic_cast<GShip*>(global_fleet_type.getCurrentObject())) {
        hyperjumping        = sh->isHyperdriving();
        current_gravity_pid = getDisplayedTurn().getCurrentUniverse().getGravityPlanetAt(sh->getPos());
    } else {
        hyperjumping = false;
        current_gravity_pid = 0;
    }
}

int32
WFleetScannerChart::lockQueryDistance(GPoint pt)
{
    // FIXME: this routine also in WShipScannerChartWidget
    if (GShip* sh = dynamic_cast<GShip*>(global_fleet_type.getCurrentObject())) {
        const int32_t dist2 = distanceSquared(sh->getPos(), pt);
        if (sh->isHyperdriving() && !host.isExactHyperjumpDistance2(dist2)) {
            return -1;
        } else {
            return dist2;
        }
    } else {
        return -1;
    }
}

void
WFleetScannerChart::drawScanner(GfxCanvas& can)
{
    // FIXME: this routine also in WShipScannerChartWidget
    if (GShip* sh = dynamic_cast<GShip*>(global_fleet_type.getCurrentObject())) {
        GChartViewport& vp = getViewport();
        GPoint wp = sh->getWaypoint();

        if (wp != getScannerTarget() && wp != getScannerOrigin()) {
            GfxPoint wp_screen = vp.scaleXY(wp);
            GfxContext ctx;
            ctx.useColorScheme(GfxStandardColorScheme::instance);
            ctx.useCanvas(can);
            ctx.setColor(COLOR_RED);
            drawLine(ctx, vp.scaleXY(getScannerOrigin()), wp_screen);
            vp.drawScannerEnd(ctx, wp_screen);
        }

        WScannerChartWidget::drawScanner(can);
    }
}

void
WFleetScannerChart::drawPre(GfxCanvas& can)
{
    if (member_pos != getCenterPosition()) {
        const GChartViewport& vp = getViewport();
        GfxContext ctx;
        ctx.useCanvas(can);
        ctx.useColorScheme(GfxStandardColorScheme::instance);
        ctx.setColor(COLOR_DARKYELLOW);

        GfxPoint a = vp.scaleXY(getCenterPosition());
        GfxPoint b = vp.scaleXY(member_pos);
        drawLine(ctx, a, b);
        drawHLine(ctx, b.x-10, b.y, b.x+10);
        drawVLine(ctx, b.x, b.y-10, b.y+10);
    }
    WScannerChartWidget::drawPre(can);
}

void
WFleetScannerChart::setMemberPosition(const GPoint pos)
{
    if (pos != this->member_pos) {
        this->member_pos = pos;
        drawWidget();
    }
}

/****************************** WFleetScreen *****************************/

/** Constructor. */
WFleetScreen::WFleetScreen()
    : WControlScreen(global_fleet_type, WScreenId::FleetScreen),
      conn_member(global_ship_type.sig_pos_changed.add(this, &WFleetScreen::onMemberChange))
{
    createTileLayout(*this, "FLEETSCREEN", global_fleet_type);
    addChart(chart);
    setKeymap(IntKeymap::getKeymapByName("FLEETSCREEN"));
}

WFleetScreen::~WFleetScreen()
{ }

void
WFleetScreen::update()
{
    if (GShip* sh = dynamic_cast<GShip*>(getType().getCurrentObject())) {
        chart.setCenterPosition(sh->getPos());
        chart.setScannerOrigin(sh->getPos());
        chart.setScannerTarget(sh->getWaypoint());
        onMemberChange();
    }
}

void
WFleetScreen::onMemberChange()
{
    if (GMapObject* mo = dynamic_cast<GMapObject*>(global_ship_type.getCurrentObject())) {
        chart.setMemberPosition(mo->getPos());
    }
}

void
WFleetScreen::setScriptContext()
{
    global_contexts.push_back(new IntShipContext(global_ship_type.getCurrent()));
}

void
doFleetScreen()
{
    WFleetScreen().runScreen();
}
