/**
  *  \file game/load.cc
  *  \brief Game Loading -- Common Code
  *
  *  This module contains routines for loading miscellaneous parts
  *  of the game data.
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

/** \defgroup Structures Data Structure Definitions
    
    Definitions of the raw structures. These are preprocessed by
    iogen.pl to generate conversion information. */

#include <algorithm>
#include "game/load.h"
#include "util/except.h"
#include "util/nls.h"
#include "game/global.h"
#include "util/console.h"
#include "game/formula.h"
#include "game/cmds.h"
#include "game/ship.h"
#include "game/planet.h"
#include "cpluslib/format.h"
#include "game/ufo.h"
#include "io/streams.h"

/** Load Ships. Loads SHIPx.DAT or the appropriate section from a RST.
    \param trn   Game Turn object
    \param s     Stream to read from
    \param count Number of ships to load
    \param where where to store data (LoadCurrent, LoadPrevious)
    \param src   Source of this information */
void
loadShips(GGameTurn& trn, Stream& s, int count, LoadUniverse where, GPlayerSet src)
{
    while (count > 0) {
        TShip ship;
        getStructureT(s, ship);
        if (ship.ship_id <= 0 || ship.ship_id > NUM_SHIPS)
            throw FileFormatException(s, format(_("Invalid ship Id #%d"), ship.ship_id));
        if (where & LoadPrevious)
            trn.getPreviousUniverse().getShip(ship.ship_id).addShipData(ship, src);
        if (where & LoadCurrent)
            trn.getCurrentUniverse().getShip(ship.ship_id).addShipData(ship, src);
        --count;
    }
}

/** Load Planets. Loads PDATAx.DAT or the appropriate section from a RST.
    \param trn   Game Turn object
    \param s     Stream to read from
    \param count Number of planets to load
    \param where where to store data (LoadCurrent, LoadPrevious)
    \param asrc  Source of this information */
void
loadPlanets(GGameTurn& trn, Stream& s, int count, LoadUniverse where, GPlayerSet src)
{
    while (count > 0) {
        TPlanet planet;
        getStructureT(s, planet);
        if (planet.planet_id <= 0 || planet.planet_id > NUM_PLANETS)
            throw FileFormatException(s, format(_("Invalid planet Id #%d"), planet.planet_id));
        if (where & LoadPrevious)
            trn.getPreviousUniverse().getPlanet(planet.planet_id).addPlanetData(planet, src);
        if (where & LoadCurrent)
            trn.getCurrentUniverse().getPlanet(planet.planet_id).addPlanetData(planet, src);
        --count;
    }
}

/** Load Starbases. Loads BDATAx.DAT or the appropriate section from a RST.
    \param trn   Game Turn object
    \param s     Stream to read from
    \param count Number of starbases to load
    \param where where to store data (LoadCurrent, LoadPrevious)
    \param src   Source of this information */
void
loadBases(GGameTurn& trn, Stream& s, int count, LoadUniverse where, GPlayerSet src)
{
    while (count > 0) {
        TStarbase base;
        getStructureT(s, base);
        if (base.base_id <= 0 || base.base_id > NUM_PLANETS)
            throw FileFormatException(s, format(_("Invalid starbase Id #%d"), base.base_id));
        if (where & LoadPrevious)
            trn.getPreviousUniverse().getPlanet(base.base_id).addBaseData(base, src);
        if (where & LoadCurrent)
            trn.getCurrentUniverse().getPlanet(base.base_id).addBaseData(base, src);
        --count;
    }
}

/** Load targets. Load TARGETx.DAT, TARGETx.EXT, or appropriate section
    from KOREx.DAT, UTILx.DAT or RST.
    \param trn   Game Turn object
    \param s     Stream to read from
    \param count Number of targets to read
    \param fmt   Format of targets (encrypted or plaintext)
    \param src   Source of this information */
void
loadTargets(GGameTurn& trn, Stream& s, int count, TargetFormat fmt, LoadUniverse where, GPlayerSet src)
{
    while (count > 0) {
        TShipTarget target;

        /* do not use getStructure(stream) here. We have to decrypt
           before recoding. */
        char buffer[TShipTarget::size];
        s.readT(buffer, sizeof buffer);
        if (fmt == TargetEncrypted)
            for (int i = 0; i < 20; ++i)
                buffer[i+14] ^= 154-i;
        getStructure(buffer, target);

        if (target.ship_id <= 0 || target.ship_id > NUM_SHIPS) {
            console.write(LOG_ERROR,
                          format(_("Invalid ship Id #%d for visual contact. Target will be ignored."), target.ship_id));
        } else {
            if (where & LoadPrevious)
                trn.getPreviousUniverse().getShip(target.ship_id).addTargetData(target, src);
            if (where & LoadCurrent)
                trn.getCurrentUniverse().getShip(target.ship_id).addTargetData(target, src);
        }
        --count;
    }
}

/** Load SHIPXY file.
    \param trn    Turn object
    \param s      Input stream
    \param avail  Number of bytes known to be available.
    \param reject Reject entries of these players.
                  This is used for allied file loading; to not wreak
                  havoc if player accidentially mixed up his files. */
void
loadShipXY(GGameTurn& trn, Stream& s, int avail, LoadUniverse where, GPlayerSet reject)
{
    static const int CHUNK_SIZE = 100;

    CTASSERT(NUM_SHIPS == 999);
    int nships = avail && avail >= 999*TShipXY::size ? 999 : 500;
    int id = 0;
    char buffer[TShipXY::size * CHUNK_SIZE];

    while (nships > 0) {
        /* Read file in chunks of CHUNK_SIZE */
        unsigned now = std::min(nships, CHUNK_SIZE);
        s.readT(buffer, now * TShipXY::size);
        for (unsigned i = 0; i < now; ++i) {
            TShipXY entry;
            getStructure(&buffer[i * TShipXY::size], entry);
            ++id;
            /* Detect bogus files made by Winplan999/Unpack999 when used
               with Host500. The SHIPXY file continues with a (mangled)
               copy of GENx.DAT which results in unlikely high coordinates.
               Only test for ship #501, to keep the risk of false positives
               low (if someone actually goes that far -- it's not forbidden
               after all). Stupid "solution" for stupid problem. */
            if (id == 501 && (entry.x < 0 || entry.x >= 0x3030 || entry.owner >= 0x2020))
                return;

            if (entry.owner > 0 && entry.owner <= NUM_OWNERS && !reject.contains(entry.owner)) {
                if (where & LoadCurrent)
                    trn.getCurrentUniverse().getShip(id).addShipXYData(entry);
                if (where & LoadPrevious)
                    trn.getPreviousUniverse().getShip(id).addShipXYData(entry);
            }
        }
        nships -= now;
    }
}

/** Load Minefields from KORE-style file.
    \param trn   Game turn object
    \param s     Stream
    \param count Number of minefields to load */
void
loadKoreMinefields(GGameTurn& trn, Stream& s, int count)
{
    /* FIXME: probably should load all at once instead of
       one-at-a-time. */
    for (int i = 1; i <= count; ++i) {
        TKoreMine mf;
        getStructureT(s, mf);
        if (mf.owner != 0) {
            /* Minefield was seen. If it is zero-size, do not let it
               come to existance if it is not already there. */
            GMinefield* pmf = (mf.radius > 0
                               ? &trn.getCurrentUniverse().getMinefields().getCreate(i, GMinefield::mfa_Scanned)
                               :  trn.getCurrentUniverse().getMinefields().get(i));
            int32 new_units = square(mf.radius);
            /* THost uses ERND(Sqrt(units)), PHost uses Trunc(Sqrt(units)).
               Therefore, actual radius is [r,r+1) in PHost, [r-.5,r+.5] for
               THost. Hence, possible unit ranges corresponding to this
               radius are (r-.5) = r-r-.25 and (r+1). */
            int32 min_units = new_units - mf.radius - 1;
            int32 max_units = square(mf.radius + 1);
            int   new_owner = mf.owner == 12 ? 7 : mf.owner;
            if (pmf) {
                bool is_same_field = (pmf->getOwner() == new_owner && pmf->getPos() == GPoint(mf.x, mf.y));
                if (is_same_field
                    && pmf->getTurnLastSeen() == trn.getTurnNumber()
                    && pmf->getUnits() >= min_units
                    && pmf->getUnits() < max_units)
                {
                    /* The minefield was already seen this turn, with
                       better information (exact unit count). */
                } else {
                    /* Update minefield. We set the type if we can deduce it
                       from the mf.owner field, or this is a whole new minefield. */
                    if (new_owner == 7 || !is_same_field)
                        pmf->setData(mf.x, mf.y, new_owner, new_units, trn.getTurnNumber(), mf.owner == 12);
                    else
                        pmf->setData(mf.x, mf.y, new_owner, new_units, trn.getTurnNumber());
                }
            }
        }
    }
}

/** Load ion storms from KOREx.DAT.
    \param trn    Game turn object
    \param s      Stream
    \param count  Number of ion storms to load */
void
loadKoreIonStorms(GGameTurn& trn, Stream& s, int count)
{
    for (int i = 1; i <= count; ++i) {
        TKoreStorm st;
        getStructureT(s, st);
        if (st.volt > 0 && st.radius > 0)
            trn.getCurrentUniverse().getIonStorm(i).addKoreData(st);
    }
}

/** Load Ufos from KOREx.DAT/UFO.HST.
    \param trn   Game turn object
    \param s     Stream
    \param first_id Id of first Ufo slot
    \param count Number of slots to read */
void
loadUfos(GGameTurn& trn, Stream& s, int first_id, int count)
{
    GUfoContainer& uc = trn.getCurrentUniverse().getUfos();
    for (int i = 0; i < count; ++i) {
        TUfo ufo;
        getStructureT(s, ufo);
        if (ufo.color != 0)
            uc.addUfoData(first_id + i, ufo);
    }
}

/** Load explosions from KOREx.DAT.
    \param trn Game turn object
    \param s   Stream
    \param count Number of slots to read */
void
loadKoreExplosions(GGameTurn& trn, Stream& s, int count)
{
    GExplosionContainer& xc = trn.getCurrentUniverse().getExplosions();
    for (int i = 0; i < count; ++i) {
        TKoreExplosion kx;
        getStructureT(s, kx);
        if (kx.x != 0 || kx.y != 0)
            xc.addExplosion(GExplosion(i+1, GPoint(kx.x, kx.y)));
    }
}

/** Extract commands from a message. This figures out the PHost
    commands from a message a player sent to himself.
    \param trn     Game turn object
    \param text    Decoded message text
    \param player  Player number
    \returns Message text without commands. Might be empty if the
    message consists entirely of commands. */
static string_t
extractCommands(GGameTurn& trn, string_t text, int player)
{
    string_t::size_type n = text.find_first_not_of(' ');
    if (n != text.npos && text[n] == '<')
        /* it's a genuine message. Don't parse. */
        return text;

    string_t s;
    do {
        string_t now = strFirst(text, "\n");
        if (GCommand::isMessageIntroducer(now))
            /* the rest is a message. */
            return s + text;
        if (GCommand* cmd = GCommand::parseCommand(now, false)) {
            trn.getCommands(player).setCommand(cmd);
        } else {
            s += now;
            s += '\n';
        }
    } while (strRemove(text, "\n"));

    string_t::size_type x = s.length();
    while (x > 0 && s[x-1] == '\n')
        --x;
    if (x < s.length())
        s.erase(x);
    return s;
}

/** Add message from message file. This decides whether the message is
    a command message or a normal message, and places it in the appropriate
    part of the game turn object (outbox, command list).
    \param trn    Game turn object
    \param text   Message text
    \param sender Sender of message
    \param recv   Receivers of message
    \note This only recognizes messages to one receiver as command
    messages. It is possible (but unlikely) that someone sends a
    message to theirselves and somone else. Our Maketurn will make
    sure that the message comes out as a real text message. However,
    with Winplan's maketurn, the message will be interpreted by
    PHost.  */
void
addMessageFromFile(GGameTurn& trn, string_t text, int sender, GPlayerSet recv)
{
    if (recv == GPlayerSet(sender)) {
        /* It's a message to us. Is it a command message? */
        text = extractCommands(trn, text, sender);
        if (text.empty())
            return;
    }

    trn.getOutbox(sender).addMessageFromFile(text, recv);
}
