/**
  *  \file game/sim-game.cc
  *  \brief Simulator / Game Interface
  *
  *  This is the "real" interface to a game, which links the simulator
  *  UI to real game data. This allows people to add ships from game
  *  to simulator and vice versa.
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/sim-game.h"

#include <algorithm>
#include "game/planet.h"
#include "game/ship.h"
#include "game/mission.h"
#include "game/cargo.h"
#include "game/global.h"
#include "game/cargotr.h"

GSimulatorRealGameInterface::GSimulatorRealGameInterface(GUniverse& univ)
    : univ(univ)
{ }

GSimulatorRealGameInterface::~GSimulatorRealGameInterface()
{ }

string_t
GSimulatorRealGameInterface::getPlanetName(int id)
{
    return univ.getPlanet(id).getName();
}

int
GSimulatorRealGameInterface::getShipOwner(int id)
{
    if (!univ.isValidShipId(id))
        return 0;

    GShip& game_ship = univ.getShip(id);
    if (game_ship.isExistant() && game_ship.isPlayed()) {
        return game_ship.getRealOwner();
    } else {
        return 0;
    }
}

bool
GSimulatorRealGameInterface::isGameAvailable()
{
    return true;
}

/** Check ship's relation to the game.
    \todo Handle targets. */
GSimulatorRealGameInterface::ShipRelation
GSimulatorRealGameInterface::getShipRelation(const GSimShip& sh)
{
    if (!univ.isValidShipId(sh.getId()))
        return ShipIsUnknown;

    GShip& game_ship = univ.getShip(sh.getId());
    if (!game_ship.isExistant() || !game_ship.isPlayed() || sh.getOwner() != game_ship.getRealOwner())
        return ShipIsUnknown;
    if (sh.getHull() != game_ship.getHullId())
        return ShipIsUnknown;

    return ShipIsPlayed;
}

bool
GSimulatorRealGameInterface::selectGameShip(const GSimShip& sh)
{
    if (!global_ship_type.isExistant(sh.getId()))
        return false;

    global_ship_type.setCurrent(sh.getId());
    return true;
}

bool
GSimulatorRealGameInterface::updateFromGame(GSimShip& sh)
{
    if (!univ.isValidShipId(sh.getId()))
        return false;

    // copyShip (is intended to) handle partial data
    return copyShip(sh, univ.getShip(sh.getId()));
}

bool
GSimulatorRealGameInterface::updateToGame(const GSimShip& sh)
{
    if (getShipRelation(sh) != ShipIsPlayed)
        return false;

    GShip& game_ship = univ.getShip(sh.getId());
    // can't modify crew, hull, beam type, beam count, torp type,
    // launcher count, bay count, engine type, Id

    // Mission
    // FIXME: check whether we're allowed to modify the mission
    if (sh.getAggressiveness() == GSimShip::agg_Kill) {
        game_ship.setMission(GMission::msn_Kill, 0, 0);
    } else {
        if (game_ship.canDoSpecial(hf_Cloak) && (sh.getFlags() & GSimShip::fl_Cloaked) != 0) {
            if (!isMissionCloaking(game_ship.getMission(), game_ship.getRealOwner()))
            {
                game_ship.setMission(GMission::msn_Cloak, 0, 0);
            }
        } else {
            if (isMissionCloaking(game_ship.getMission(), game_ship.getRealOwner())
                || game_ship.getMission() == GMission::msn_Kill)
            {
                game_ship.setMission(0, 0, 0);
            }

            if (sh.getInterceptId() != 0) {
                // Take over Intercept mission if allowed
                int i = sh.getInterceptId();
                if (univ.isValidShipId(i)
                    && univ.getShip(i).isExistant()
                    && univ.getShip(i).isReliablyVisible())
                {
                    game_ship.setMission(GMission::msn_Intercept, i, 0);
                }
            }
        }
    }

    // PE
    if (sh.getAggressiveness() > 0 && sh.getAggressiveness() <= NUM_PLAYERS)
        game_ship.setPE(sh.getAggressiveness());
    else
        game_ship.setPE(0);

    // FCode, Name
    game_ship.setFCode(sh.getFCode());
    game_ship.setName(sh.getName());

    // Ammo
    // FIXME: in case we handle mkt/lfm someday, we would have to revert it here
    int pid;
    if (sh.getAmmo() != game_ship.getAmmo()
        && (pid = game_ship.getOrbit()) != 0)
    {
        GPlanet& game_planet = univ.getPlanet(pid);
        if (game_planet.getOwner() == game_ship.getOwner()
            && game_planet.isPlayed()
            && game_planet.hasPlanetData())
        {
            GCargoType type;
            if (sh.getTorpLauncherCount() != 0)
                type = getCargoTypeFromTorpType(sh.getTorpType());
            else if (sh.getBayCount() != 0)
                type = el_Fighters;
            else
                type = el_Neutronium;

            if (type != el_Neutronium) {
                // use regular cargo transfer
                GCargoTransfer ct;
                ct.addNewContainer(new GPlanetTransfer(game_planet));
                ct.addNewContainer(new GShipTransfer(game_ship));
                ct.move(type, sh.getAmmo() - game_ship.getAmmo(), false, true, false);
                ct.doCommit();
            }
        }
    }

    return true;
}

bool
GSimulatorRealGameInterface::isGamePlanet(const GSimPlanet& pl)
{
    if (!univ.isValidPlanetId(pl.getId()))
        return false;

    GPlanet& game_planet = univ.getPlanet(pl.getId());
    return game_planet.isExistant()
        && game_planet.hasPlanetData()
        && game_planet.isPlayed()
        && game_planet.getOwner() == pl.getOwner();
}

bool
GSimulatorRealGameInterface::updateFromGame(GSimPlanet& pl)
{
    if (!univ.isValidPlanetId(pl.getId()))
        return false;

    // copyPlanet (is intended to) handle partial data
    return copyPlanet(pl, univ.getPlanet(pl.getId()));
}

bool
GSimulatorRealGameInterface::updateToGame(const GSimPlanet& pl)
{
    // FIXME: this function isn't yet used by the UI, hence not yet implemented
    (void) pl;
    return false;
}

/*************************** Low-Level Routines **************************/

/** Assign hull function bits from game ship.
    \param game_ship [in] Game ship to copy from
    \param sh        [out] Simulated ship
    \param simf      [in] Function to consider */
static void
setHullFunction(const GShip& game_ship, GSimShip& sh, const GSimFunction simf)
{
    bool ship_can_do = game_ship.canDoSpecial(GSimShip::simf_to_hullfunc[simf]);
    bool sim_can_do  = sh.hasImpliedHullFunction(simf);
    if (ship_can_do != sim_can_do) {
        sh.setFlags(sh.getFlags() | GSimShip::simf_to_set_bit[simf]);
        if (ship_can_do)
            sh.setFlags(sh.getFlags() | GSimShip::simf_to_value_bit[simf]);
    } else {
        sh.setFlags(sh.getFlags() & ~GSimShip::simf_to_set_bit[simf]);
    }
}

static int16
chooseKnown(mp16_t shipval, int16 specval)
{
    return shipval.isKnown() ? shipval : specval;
}

bool
GSimulatorRealGameInterface::copyShip(GSimShip& sh, const GShip& game_ship)
{
    if (game_ship.hasShipData()) {
        sh.setCrew(game_ship.getCrew());
        sh.setHull(game_ship.getHullId());
        sh.setBeamType(game_ship.getBeamType());
        sh.setBeamCount(game_ship.getBeamCount());
        sh.setTorpType(game_ship.getTorpType());
        sh.setTorpLauncherCount(game_ship.getTorpLauncherCount());
        sh.setBayCount(game_ship.getBayCount());
        sh.setAmmo(game_ship.getAmmo());
        sh.setEngineType(game_ship.getEngine());

        if (game_ship.getCargo(el_Neutronium) == 0)
            sh.setAggressiveness(sh.agg_NoFuel);
        else if (game_ship.getMission() == GMission::msn_Kill)
            sh.setAggressiveness(sh.agg_Kill);
        else
            sh.setAggressiveness(game_ship.getPE());

        sh.setId(game_ship.getId());
        sh.setName(game_ship.getName());
        sh.setFCode(game_ship.getFCode());
        sh.setDamage(std::max(int(game_ship.getDamage() - game_ship.getCargo(el_Supplies) / 5), 0));
        sh.setShield(std::max(int(100 - sh.getDamage()), 0));
        sh.setOwner(game_ship.getRealOwner());
        sh.setExperienceLevel(0); // FIXME

        sh.setFlags(0);
        if (isMissionCloaking(game_ship.getMission(), game_ship.getRealOwner())
            && game_ship.canDoSpecial(hf_Cloak)
            && game_ship.getCargo(el_Neutronium) > 0)
        {
            sh.setFlags(sh.getFlags() | GSimShip::fl_Cloaked);
        }

        if (game_ship.getMission() == GMission::msn_Intercept && game_ship.canDoSpecial(hf_Cloak)) {
            sh.setInterceptId(game_ship.getIntId());
        } else {
            sh.setInterceptId(0);
        }

        setHullFunction(game_ship, sh, sf_FullWeaponry);
        setHullFunction(game_ship, sh, sf_PlanetImmunity);
        setHullFunction(game_ship, sh, sf_Commander);

        // FIXME: FLAK rating override

        return true;
    } else if (game_ship.getHullId() > 0) {
        // Hull is known, so this is a basic enemy contact
        const GHull& h = ::getHull(game_ship.getHullId());
        sh.setCrew(chooseKnown(game_ship.getCrew(), h.getCrew()));
        sh.setHull(game_ship.getHullId());
        sh.setBeamType(chooseKnown(game_ship.getBeamType(), h.getMaxBeams() > 0 ? NUM_BEAMS : 0));
        sh.setBeamCount(chooseKnown(game_ship.getBeamCount(), h.getMaxBeams()));
        sh.setTorpType(chooseKnown(game_ship.getTorpType(), h.getMaxLaunchers() > 0 ? NUM_TORPS : 0));
        sh.setTorpLauncherCount(chooseKnown(game_ship.getTorpLauncherCount(), h.getMaxLaunchers()));
        sh.setBayCount(chooseKnown(game_ship.getBayCount(), h.getNumBays()));

        // GShip::getAmmo() currently can only be called for ships that have a known
        // weapon type.
        if (game_ship.getTorpType().isKnown() && game_ship.getBayCount().isKnown() && game_ship.getAmmo() >= 0)
            sh.setAmmo(game_ship.getAmmo());
        else
            sh.setAmmo(h.getCargo());

        sh.setEngineType(chooseKnown(game_ship.getEngine(), NUM_ENGINES));
        sh.setAggressiveness(sh.agg_Kill);
        sh.setId(game_ship.getId());
        sh.setName(game_ship.getName());
        sh.setFCode("???");
        sh.setDamage(0);
        sh.setShield(100);
        sh.setOwner(game_ship.getRealOwner());
        sh.setExperienceLevel(0); // FIXME

        sh.setFlags(0);
        setHullFunction(game_ship, sh, sf_FullWeaponry);
        setHullFunction(game_ship, sh, sf_PlanetImmunity);
        setHullFunction(game_ship, sh, sf_Commander);
        return true;
    } else {
        // FIXME: handle partial data
        return false;
    }
}

bool
GSimulatorRealGameInterface::copyPlanet(GSimPlanet& pl, const GPlanet& game_planet)
{
    if (game_planet.hasPlanetData()) {
        pl.setId(game_planet.getId());
        pl.setName(game_planet.getName());
        pl.setFCode(game_planet.getFCode());
        pl.setDefense(game_planet.getStructures(str_Defense));
        pl.setDamage(0);
        pl.setShield(0);
        pl.setOwner(game_planet.getOwner());

        if (game_planet.hasBaseData()) {
            pl.setBaseDefense(game_planet.getStructures(str_BaseDefense));
            pl.setBaseBeamTech(game_planet.getBaseTech(tl_Beam));
            pl.setBaseTorpTech(game_planet.getBaseTech(tl_Torpedo));
            pl.setBaseFighters(game_planet.getCargo(el_Fighters));
            for (int i = 1; i <= NUM_TORPS; ++i)
                pl.setBaseTorps(i, game_planet.getCargo(getCargoTypeFromTorpType(i)));
            pl.setBaseDamage(game_planet.getBaseDamage());
        } else {
            pl.setBaseDefense(0);
            pl.setBaseBeamTech(0);
            pl.setBaseTorpTech(0);
            pl.setBaseFighters(0);
            for (int i = 1; i <= NUM_TORPS; ++i)
                pl.setBaseTorps(i, 0);
            pl.setBaseDamage(0);
        }

        pl.setExperienceLevel(0); // FIXME
        pl.setFlags(0);

        // FIXME: FLAK rating override

        return true;
    } else {
        // FIXME: handle partial data
        return false;
    }
}
