/**
  *  \file ui/listbox.h
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UI_LISTBOX_H
#define PCC_V2_UI_LISTBOX_H

#include <list>
#include <utility>
#include "ui/ui.h"
#include "cpluslib/sigslot.h"
#include "ui/uiglobal.h"
#include "util/strlist.h"

class GfxContext;

class UIScrollable : public UIWidget {
    int pos, top, size, page;
 protected:
    void setCurrent(int i)  { pos  = i; }
    void setTop(int i)      { top  = i; }
    void setSize(int i)     { size = i; }
    void setPageSize(int i) { page = i; }
    void adjustPosition();
 public:
    UIScrollable(int id);
    ~UIScrollable();
    virtual void scrollTo(int pos, int corr) = 0;
    virtual bool isAccessible(int pos);
    int getCurrent() const  { return pos; }
    int getTop() const      { return top; }
    int getSize() const     { return size; }
    int getPageSize() const { return page; }

    Signal1<int> sig_moved;
};

class UIListbox : public UIScrollable {
    int lines, width;           /* user preferences */
    int item_height;
    int header_height;
    bool mdown;
    bool have_menu_pos;
    GfxPoint menu_pos;
 protected:
    void setItemHeight(int h) { item_height = h; }
    void setHeaderHeight(int h) { header_height = h; }
 public:
    UIListbox(int id, int lines, int width, int items);

    void scrollTo(int pos, int corr);
    void drawContent(GfxCanvas& can);
    void onStateChange(int state, bool enable);
    bool handleEvent(const UIEvent& e, bool second_pass);
    bool handleCommand(const UICommand& cmd, UIBaseWidget& sender);
    void updateCurrent();
    void onResize(GfxRect oldsize, bool by_layout);
    void setDefaultHeight();

    void getLayoutInfo(LayoutInfo& info);

    int getItemHeight() const { return item_height; }
    int getHeaderHeight() const { return header_height; }

    virtual void drawPart(GfxCanvas& can, int from, int to) = 0;
    virtual void onItemClick(int item, int x);

    Signal1<int> sig_dblclick, sig_menu;

    enum {
        lb_Menu         = 2*st_Last,
        lb_NoPageKeys   = 4*st_Last,
        lb_KeyboardMenu = 8*st_Last
    };

    bool       doStandardListbox(string_t title, int help_id=0);
    bool       doStandardMenu(int x, int y);
    bool       doStandardMenu(UIBaseWidget& ref_widget);

    enum Style {
        /** High contrast. Use widget skin. Color will be set to a skin color depending
            on item state; be careful when using other colors. */
        HighContrastStyle,
        /** Color style. Uses fixed set of colors. All UIColor can be used, tc_Static
            will be preselected. */
        ColorStyle,
        /** Like ColorStyle, but without cursor. No highlighting for the selected item
            will be done. */
        CursorlessColorStyle
    };

    void prepareEntry(GfxContext& ctx, Style style, int line, GfxRect* area_out);
};

typedef StringList UIStringList;

class UIStandardListbox : public UIListbox {
    const UIStringList& data;
 public:
    UIStandardListbox(int id, int lines, int width, const UIStringList& data);
    ~UIStandardListbox();

    void       drawPart(GfxCanvas& can, int from, int to);
    int        getKey() const
        { return data.getItem(getCurrent())->first; }
    void       scrollToKey(int key);

    bool       handleEvent(const UIEvent& e, bool second_pass);
};

class UIScrollbar : public UIWidget {
    UIScrollable&     scrl;
    SignalConnection  conn;
    int               bf[2];

    void sync();
    void scrolled(int);
    void setBF(int id, int flags);
 public:
    UIScrollbar(UIScrollable& scrl);
    ~UIScrollbar();

    void drawContent(GfxCanvas& can);
    void getLayoutInfo(LayoutInfo& info);
    bool handleEvent(const UIEvent& e, bool second_pass);
    void onStateChange(int flags, bool enable);
};

#endif
